(function(){
	var
		PRELOAD_INTERVAL, PRELOAD_CACHE = {},
		PRELOAD_LOAD = [], PRELOAD_ERROR = [], PRELOAD_READY = [];


	window.preloadLoad = function(Callback){
		if(typeof(Callback) == 'function'){
			PRELOAD_LOAD.push(Callback);
		}
	};

	window.preloadError = function(Callback){
		if(typeof(Callback) == 'function'){
			PRELOAD_ERROR.push(Callback);
		}
	};

	window.preloadReady = function(Callback){
		if(typeof(Callback) == 'function'){
			PRELOAD_READY.push(Callback);
		}
	};

	window.preloadReady(function(){
		var preload = document.querySelectorAll('.preload');

		if(preload instanceof NodeList){
			preload.forEach(function(Element){
				Element.style.removeProperty('display');
			});
		}
	});


	function linkRegister(link){
		if(!PRELOAD_CACHE.hasOwnProperty(link)){
			PRELOAD_CACHE[link] = false;
		}
	}

	function linkLoad(link){
		PRELOAD_CACHE[link] = true;

		PRELOAD_LOAD.forEach(function(Callback){
			try{
				Callback(link);
			}catch(exception){
				console.error(exception.message);
			}
		});
	}

	function linkError(link){
		PRELOAD_CACHE[link] = true;

		PRELOAD_ERROR.forEach(function(Callback){
			try{
				Callback(link);
			}catch(exception){
				console.error(exception.message);
			}
		});
	}

	function linkReady(){
		PRELOAD_READY.forEach(function(Callback){
			try{
				Callback();
			}catch(exception){
				console.error(exception.message);
			}
		});
	}


	// Overwrite document.write after document loaded
	function overwriteFunctions(){
		document.write = function(hAlloc){ document.currentScript.parentNode.innerHTML += hAlloc; };
		document.writeln = function(hAlloc){ document.currentScript.parentNode.innerHTML += hAlloc + '\n'; };
	};


	function copyAttributes(Source, Destination, ExcludeAttributes){
		var
			attribute, key, value,
			attributes = Source.attributes,
			index, count = attributes.length;

		for(index = 0; index < count; index++){
			attribute = attributes[index];
			key = (typeof(attribute.name) == 'string' ? attribute.name.trim().toLowerCase() : '');
			value = (typeof(attribute.value) == 'string' ? attribute.value : '');

			if((key != '') && !ExcludeAttributes.includes(key)){
				Destination.setAttribute(key, value);
			}
		}
	};


	function modulePreload(){
		document
			.querySelectorAll('link[href][rel=modulepreload]')
			.forEach(function(Link){
				var
					href = Link.getAttribute('href'),
					element = document.createElement('script');


				linkRegister(href);

				element.setAttribute('type', 'module');
				element.setAttribute('src', href);

				element.onload = function(){ linkLoad(href); };
				element.onerror = function(){ linkError(href); };

				copyAttributes(Link, element, ['href', 'rel', 'type', 'src', 'onload', 'onerror']);

				document.head.appendChild(element);
				Link.remove();
			});
	};

	function preloadStylesheet(){
		document
			.querySelectorAll('link[href][rel=preload][as=style]')
			.forEach(function(Link){
				var
					href = Link.getAttribute('href'),
					element = document.createElement('link');


				linkRegister(href);

				element.setAttribute('rel', 'stylesheet');
				element.setAttribute('href', href);

				element.onload = function(){ linkLoad(href); };
				element.onerror = function(){ linkError(href); };

				copyAttributes(Link, element, ['href', 'rel', 'as', 'onload', 'onerror']);

				document.head.appendChild(element);
				Link.remove();
			});
	};

	function preloadScriptWithAttribute(){
		document
			.querySelectorAll(
				'link[href][rel=preload][as=script][async], ' +
				'link[href][rel=preload][as=script][defer]'
			)
			.forEach(function(Link){
				var
					href = Link.getAttribute('href'),
					element = document.createElement('script');


				linkRegister(href);

				element.setAttribute('type', 'text/javascript');
				element.setAttribute('src', href);

				element.onload = function(){ linkLoad(href); };
				element.onerror = function(){ linkError(href); };

				copyAttributes(Link, element, ['href', 'rel', 'as', 'type', 'src', 'onload', 'onerror']);

				document.head.appendChild(element);
				Link.remove();
			});
	};


	function preloadScriptInQuery(){
		var Link = document.querySelector('link[href][rel=preload][as=script]');

		if(Link instanceof HTMLLinkElement){
			var
				href = Link.getAttribute('href'),
				element = document.createElement('script');


			linkRegister(href);

			element.setAttribute('type', 'text/javascript');
			element.setAttribute('src', href);

			element.onload = function(){ linkLoad(href); preloadScriptInQuery(); };
			element.onerror = function(){ linkError(href); preloadScriptInQuery(); };

			copyAttributes(Link, element, ['href', 'rel', 'as', 'type', 'src', 'onload', 'onerror']);

			document.head.appendChild(element);
			Link.remove();
		}
	};

	function preloadEvent(){
		if(Object.keys(PRELOAD_CACHE).length > 0){
			PRELOAD_INTERVAL = setInterval(function(){
				var is_ready = true, key;

				for(key in PRELOAD_CACHE){
					if(!PRELOAD_CACHE[key]){
						is_ready = false;
					}
				}

				if(is_ready){
					clearInterval(PRELOAD_INTERVAL);

					linkReady();
				}
			}, 100);
		}else{
			linkReady();
		}
	};


	overwriteFunctions();

	modulePreload();
	preloadStylesheet();
	preloadScriptWithAttribute();
	preloadScriptInQuery();

	preloadEvent();
})();
